{-# LANGUAGE TemplateHaskell #-}

module ISTEX.Client where

import Control.Applicative ((<|>))
import Control.Lens qualified as L
import Data.Aeson
import Data.Proxy (Proxy(..))
import Data.Text qualified as T
import GHC.Generics
import Servant.API
import Servant.Client
import Text.Read (readEither)

data Author = Author
  { _author_name         :: T.Text
  , _author_affiliations :: [T.Text]
  } deriving (Show, Generic)
L.makeLenses ''Author

instance FromJSON Author where
  parseJSON  = withObject "Author" $ \o -> do
    _author_name <- o .: "name"
    _author_affiliations <- o .: "affiliations" <|> pure []
    pure $ Author { .. }

data Source = Source
  { _source_title           :: Maybe T.Text
  -- , _source_authors         :: [Author]
  , _source_publicationDate :: Maybe Int
  } deriving (Show, Generic)
L.makeLenses ''Source

parsePubDate :: Maybe T.Text -> Maybe Int
parsePubDate Nothing = Nothing
parsePubDate (Just pubDate) = 
  case ePubDate of
    Left err -> error $ T.unpack $ "[parseJSON] Cannot parse publicationDate (Int): " <> pubDate <> "(" <> (T.pack err) <> ")"
    Right pubDate' -> Just pubDate'
  where
    ePubDate = readEither (T.unpack pubDate) :: Either String Int

instance FromJSON Source where
  parseJSON = withObject "Source" $ \o -> do
    _source_title     <- o .:? "title"
    -- _source_authors <- o .:? "author"
    mPubDate        <- o .:? "publicationDate"

    pure $ Source { _source_publicationDate = parsePubDate mPubDate
                  , .. }

data Document = Document
  { _document_id              :: T.Text
  , _document_title           :: Maybe T.Text
  , _document_authors         :: [Author]
  , _document_abstract        :: Maybe T.Text
  , _document_publicationDate :: Maybe Int
  , _document_sources         :: Maybe Source
  } deriving (Show, Generic)
L.makeLenses ''Document

instance FromJSON Document where
  parseJSON = withObject "Document" $ \o -> do
    _document_id       <- o .: "id" <|> o .: "_id"
    _document_title    <- o .:? "title"
    _document_authors  <- o .: "author" <|> pure []
    _document_abstract <- o .:? "abstract"
    _document_sources  <- o .:? "host"
    mPubDate           <- o .:? "publicationDate"
    
    pure $ Document { _document_publicationDate = parsePubDate mPubDate, .. }

data Documents = Documents
  { _documents_total :: Int
  , _documents_hits  :: [Document]
  } deriving (Show, Generic)
L.makeLenses ''Documents
instance FromJSON Documents where
  parseJSON = withObject "Documents" $ \o -> do
    _documents_total <- o .: "total"
    _documents_hits <- o .: "hits"
    pure $ Documents { .. }
instance Semigroup Documents where
  (<>) (Documents { _documents_total
                  , _documents_hits = d1 })
       (Documents { _documents_hits = d2 }) = Documents { _documents_total
                                                        , _documents_hits = d1 <> d2 }

instance Monoid Documents where
  mempty = Documents { _documents_total = 0
                     , _documents_hits = [] }

data ScrollResponse = ScrollResponse
  { _scroll_documents           :: Documents
  , _scroll_noMoreScrollResults :: Maybe Bool
  , _scroll_nextScrollURI       :: Maybe T.Text
  , _scroll_scrollId            :: Maybe T.Text }
L.makeLenses ''ScrollResponse

instance FromJSON ScrollResponse where
  parseJSON = withObject "ScrollResponse" $ \o -> do
    _documents_total            <- o .: "total"
    _documents_hits             <- o .: "hits"
    _scroll_noMoreScrollResults <- o .:? "noMoreScrollResults"
    _scroll_nextScrollURI       <- o .:? "nextScrollURI"
    _scroll_scrollId            <- o .:? "scrollId"
    pure $ ScrollResponse { _scroll_documents = Documents { .. }
                          , .. }

type ISTEXAPI = Search

type Search = QueryParam "defaultOperator" T.Text
           :> QueryParam "output" T.Text
           :> QueryParam "size" Int
           :> QueryParam "q" T.Text
           :> Get '[JSON] Documents

istexProxy :: Proxy (ISTEXAPI)
istexProxy = Proxy

search :: Maybe T.Text -> Maybe Int -> Maybe T.Text -> ClientM Documents
search = client istexProxy (Just "AND")

type SearchScroll = QueryParam "defaultOperator" T.Text
                 :> QueryParam "output" T.Text
                 :> QueryParam "scroll" T.Text
                 :> QueryParam "q" T.Text
                 :> QueryParam "scrollId" T.Text
                 :> Get '[JSON] ScrollResponse

istexProxyScroll :: Proxy SearchScroll
istexProxyScroll = Proxy

searchScroll :: Maybe T.Text -> T.Text -> Maybe T.Text -> Maybe T.Text -> ClientM ScrollResponse
searchScroll output scroll = client istexProxyScroll (Just "AND") output (Just scroll)
